//
//  Winmail.m
//  TNEFsEnough3
//
//  Created by Josh Jacob on 5/27/11.
//  Copyright 2011 Josh Jacob. All rights reserved.
//

#import "Winmail.h"
#import "LibytnefFileItem.h"
#import "FileTableViewDataSource.h"
#import "ytnef.h"
#import "PrintTNEF.h"
#import "LibytnefHelper.h"

@implementation Winmail

@synthesize tnefHelper;
@synthesize fileDataSource;
@synthesize fieldFrom;
@synthesize fieldSubject;
@synthesize fieldDate;
@synthesize tableFiles;
@synthesize fieldBody;
@synthesize fieldBodyScrollView;
@synthesize webview;
@synthesize webViewBox;
@synthesize printView;
@synthesize printFieldFrom;
@synthesize printFieldSubject;
@synthesize printFieldDate;
@synthesize printFieldBody;
@synthesize printWebview;
@synthesize exportFile;
@synthesize exportDir;
@synthesize exportArray;
@synthesize outputStream;
@synthesize savingSheet;
@synthesize savingSheetLabel;
@synthesize savingSheetProgress;

- (id)init
{
    self = [super init];
    if (self) {
        // Add your subclass-specific initialization here.
        // If an error occurs here, send a [self release] message and return nil.
        
        exportArray = [[NSMutableArray alloc] initWithCapacity: 1];
    }
    return self;
}

- (void)awakeFromNib
{
    [self.tableFiles setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
    [self.tableFiles registerForDraggedTypes:[NSArray arrayWithObject: NSFilesPromisePboardType] ];
}

- (void) dealloc
{
    
    // release datasource
    self.fileDataSource = nil;
    self.exportArray = nil;
    self.outputStream = nil;
    
    [super dealloc];
}

- (NSString *)windowNibName
{
    // Override returning the nib file name of the document
    // If you need to use a subclass of NSWindowController or if your document supports multiple NSWindowControllers, you should remove this method and override -makeWindowControllers instead.
    return @"Winmail";
}

#pragma mark - NSDocument

- (void)runPageLayout:(id)sender
{
    [NSApp runPageLayout: sender];
}

//- (BOOL)isInViewingMode
//{
//    return YES;
//}

- (void)windowControllerDidLoadNib:(NSWindowController *)aController
{
    [super windowControllerDidLoadNib:aController];
    // Add any code here that needs to be executed once the windowController has loaded the document's window.
    
    [self.tableFiles setDataSource: self.fileDataSource];
    
    if (tnefHelper.bodyIsHTML) 
    {
        [[self.webview mainFrame] loadHTMLString:tnefHelper.bodyHTML baseURL:nil];
        [[self.printWebview mainFrame] loadHTMLString:tnefHelper.bodyHTML baseURL:nil];
        [self.fieldBodyScrollView setHidden:YES];
        [self.webViewBox setHidden:NO];
        [self.webview setHidden:NO];
    }
}

- (NSData *)dataOfType:(NSString *)typeName error:(NSError **)outError {
    /*
     Insert code here to write your document to data of the specified type. If outError != NULL, ensure that you create and set an appropriate error when returning nil.
    You can also choose to override -fileWrapperOfType:error:, -writeToURL:ofType:error:, or -writeToURL:ofType:forSaveOperation:originalContentsURL:error: instead.
    */
    if (outError) {
        *outError = [NSError errorWithDomain:NSOSStatusErrorDomain code:unimpErr userInfo:NULL];
    }
    return nil;
}

- (BOOL)readFromURL:(NSURL *)absoluteURL ofType:(NSString *)typeName error:(NSError **)outError
{
    LibytnefHelper * helper = [[LibytnefHelper alloc] initWithFile:absoluteURL];
    self.tnefHelper = helper;
    [helper release];
    
    // test parse result
    if (helper.parseResult < 0) {
        DLog(@"ERROR processing file\n");
        NSDictionary * errorUserInfo = nil;
        /*
             #define YTNEF_CANNOT_INIT_DATA -1
             #define YTNEF_NOT_TNEF_STREAM  -2
             #define YTNEF_ERROR_READING_DATA -3
             #define YTNEF_NO_KEY -4
             #define YTNEF_BAD_CHECKSUM -5
             #define YTNEF_ERROR_IN_HANDLER -6
             #define YTNEF_UNKNOWN_PROPERTY -7
             #define YTNEF_INCORRECT_SETUP -8
         */
        switch (helper.parseResult) {
            case YTNEF_CANNOT_INIT_DATA:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Cannot initialize data.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_NOT_TNEF_STREAM:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"This is not a TNEF file.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_ERROR_READING_DATA:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Error reading file.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_NO_KEY:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"No TNEF key found.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_BAD_CHECKSUM:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Bad checksum.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_ERROR_IN_HANDLER:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Error in handler.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_UNKNOWN_PROPERTY:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Unknown property.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_INCORRECT_SETUP:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Incorrect setup.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            default:
                break;
        }
        NSError * err = [NSError errorWithDomain:@"TNEF's Enough Error Domain." code: helper.parseResult userInfo:errorUserInfo];
        *outError = err;
        return NO;
    }
    
    fileDataSource = [[FileTableViewDataSource alloc] init];
    [fileDataSource.fileArray addObjectsFromArray: tnefHelper.fileItems];
    
    return YES;
}

- (BOOL)readFromData:(NSData *)data ofType:(NSString *)typeName error:(NSError **)outError {
    /*
    Insert code here to read your document from the given data of the specified type. If outError != NULL, ensure that you create and set an appropriate error when returning NO.
    You can also choose to override -readFromFileWrapper:ofType:error: or -readFromURL:ofType:error: instead.
    */
    if (outError) {
        *outError = [NSError errorWithDomain:NSOSStatusErrorDomain code:unimpErr userInfo:NULL];
    }
    return YES;
}

- (BOOL)validateUserInterfaceItem:(id <NSValidatedUserInterfaceItem>)anItem
{
    SEL theAction = [anItem action];
    
    if (theAction == @selector(exportSelectedAttachment:))
    {
        if ([self.tableFiles numberOfSelectedRows] > 0)
        {
            return YES;
        }
        return NO;
    }
    
    if (theAction == @selector(exportAll:))
    {
        if ([self.tableFiles numberOfRows] > 0)
        {
            return YES;
        }
        return NO;
    }
    
    return [super validateUserInterfaceItem:anItem];
}

- (NSPrintOperation *)printOperationWithSettings:(NSDictionary *)printSettings error:(NSError **)outError
{
    NSPrintInfo *printInfo = [self printInfo];
    [printInfo setHorizontallyCentered:NO];
    [printInfo setVerticallyCentered:NO];
    [printInfo setHorizontalPagination: NSFitPagination];
    [printInfo setVerticalPagination: NSAutoPagination];
    
    if (tnefHelper.bodyIsHTML) {
        
        // [[self.printWebview mainFrame] loadHTMLString:tnefHelper.bodyHTML baseURL:nil];
        
        [[[self.printWebview mainFrame] frameView] setAllowsScrolling: NO];
        
        NSRect webRect = [[[[self.printWebview mainFrame] frameView] documentView] frame];
        webRect.size.height += 85;
        
        [self.printView setFrame:webRect];
        
        [self.printWebview setHidden:NO];
        [self.printFieldBody setHidden:YES];
        
    } else {
        
        [self.printFieldBody insertText: self.tnefHelper.body];
        
        NSRect bodyRect = self.printFieldBody.frame;
        bodyRect.size.height += 85;
        
        [self.printView setFrame:bodyRect];
        
        [self.printWebview setHidden:YES];
        [self.printFieldBody setHidden:NO];
    }
    
    NSPrintOperation *printOp = [NSPrintOperation printOperationWithView:self.printView printInfo:printInfo];
    return printOp;
}

#pragma mark - IBActions

- (IBAction) exportSelectedAttachment: (id) sender
{
    // test if more than one selected
    if ([self.tableFiles numberOfSelectedRows] > 1) {
        [self exportSet: [self.fileDataSource.fileArray objectsAtIndexes:[self.tableFiles selectedRowIndexes]]];
        return;
    }
    
    // get filename
    LibytnefFileItem * filetableItem = [self.fileDataSource.fileArray objectAtIndex:[self.tableFiles selectedRow]];
    NSString * filename = filetableItem.fileName;
    
    NSWindow * window = [[[self windowControllers] objectAtIndex:0] window];
        
    // Set the default name for the file and show the panel.
    NSSavePanel * panel = [NSSavePanel savePanel];
    [panel setNameFieldStringValue: filename];
    [panel beginSheetModalForWindow:window completionHandler:^(NSInteger result) {
        if (result == NSFileHandlingPanelOKButton)
        {
            // setup single file export
            exportOne = YES;
            self.exportFile = [panel URL];
            [self.exportArray removeAllObjects];
            [self.exportArray addObject:filetableItem];
            DLog(@"Will export to file: %@", self.exportFile);
            
            [self performSelectorOnMainThread:@selector(kickOffExport) withObject:nil waitUntilDone:NO];
        }
    }];
    
}

- (void) kickOffExport
{
    // show saving sheet
    [self.savingSheetProgress startAnimation: self];
    [NSApp beginSheet:self.savingSheet 
       modalForWindow:[self windowForSheet] 
        modalDelegate:self 
       didEndSelector:nil 
          contextInfo:nil];
    
    // display filename
    LibytnefFileItem * filetableItem = [self.exportArray lastObject];
    NSString * filename = filetableItem.fileName;
    NSString * newLabel = [NSString stringWithFormat:@"Saving file \"%@\"...", filename];
    [self.savingSheetLabel setStringValue:newLabel];
    [self.exportArray removeLastObject];
    
    // create output stream
    self.outputStream = [[NSOutputStream alloc] initWithURL:self.exportFile append:NO];
    [self.outputStream setDelegate:self];
    [self.outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
    [self.outputStream open];
}

- (IBAction) exportAll: (id) sender
{
    [self exportSet: self.fileDataSource.fileArray];
}

- (void) exportSet: (NSArray *) rows
{
    // get the main window for the document.
    NSWindow* window = [[[self windowControllers] objectAtIndex:0] window];
    
    // create and configure the panel.
    NSOpenPanel* panel = [NSOpenPanel openPanel];
    [panel setCanChooseDirectories:YES];
    [panel setCanChooseFiles:NO];
    [panel setAllowsMultipleSelection:NO];
    [panel setCanCreateDirectories:YES];
    [panel setMessage:@"Select export directory."];
    
    // display the panel attached to the document's window.
    [panel beginSheetModalForWindow:window completionHandler:^(NSInteger result){
        if (result == NSFileHandlingPanelOKButton) {
            NSArray* urls = [panel URLs];
            
            // setup export to dir
            exportOne = NO;
            self.exportDir = [urls objectAtIndex:0];;
            DLog(@"Will export to dir: %@", self.exportDir);
            
            // add all file to array
            for (LibytnefFileItem * fileTableItem in rows) {
                [self.exportArray addObject:fileTableItem];
            }
            
            [self performSelectorOnMainThread:@selector(kickOffExportAll) withObject:nil waitUntilDone:NO];
        }
        
    }];
}

- (void) kickOffExportAll
{
    // show saving sheet
    [self.savingSheetProgress startAnimation: self];
    [NSApp beginSheet:self.savingSheet 
       modalForWindow:[self windowForSheet] 
        modalDelegate:self 
       didEndSelector:nil 
          contextInfo:nil];
    
    [self kickOffExportAllNext];
}

- (void) kickOffExportAllNext
{
    // NSFileManager * fileManager = [[NSFileManager alloc] init];

    // get filename
    LibytnefFileItem * filetableItem = [self.exportArray lastObject];
    NSString * filename = filetableItem.fileName;
    NSString * newLabel = [NSString stringWithFormat:@"Saving file \"%@\"...", filename];
    [self.savingSheetLabel setStringValue:newLabel];
    
    // get unique name
    NSString * destFile  = [FileTableViewDataSource GenerateUniqueFileNameAtPath: [self.exportDir path] WithBase:[filename stringByDeletingPathExtension] Extension:[filename pathExtension]];
    NSURL * destFileUrl = [NSURL fileURLWithPath:destFile];
    DLog(@"Saving file %@", destFile);
    
    // create output stream
    self.outputStream = [[NSOutputStream alloc] initWithURL:destFileUrl append:NO];
    [self.outputStream setDelegate:self];
    [self.outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
    [self.outputStream open];
    
    // [fileManager release];
}

#pragma mark - NSStreamDelegate

- (void)stream:(NSStream *)theStream handleEvent:(NSStreamEvent)streamEvent
{
    if (theStream == self.outputStream)
    {
        NSOutputStream * theOutputStream = (NSOutputStream *) theStream;
        
        switch(streamEvent)
        {
            case NSStreamEventHasSpaceAvailable:
            {
                // get reference to export file and attachment
                LibytnefFileItem * filetableItem;
                if (exportOne == YES) {
                    filetableItem = [self.fileDataSource.fileArray objectAtIndex:[self.tableFiles selectedRow]];
                } else {
                    filetableItem = (LibytnefFileItem *) [self.exportArray lastObject];
                }
                Attachment * attachment = filetableItem.attachment;
                
                // write out data
                NSNumber * size = [NSNumber numberWithInt:attachment->FileData.size];
                uint8_t * data = (uint8_t *) attachment->FileData.data;
                NSInteger len = [theOutputStream write: data maxLength: [size unsignedLongValue]];
                DLog(@"Written length: %ld", len);
                
                // close up current stream
                [theOutputStream close];
                [theOutputStream removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
                [theOutputStream release];
                self.outputStream = nil;
                
                // remove last object; look for more
                if (exportOne == NO) {
                    [self.exportArray removeLastObject];
                }
                if ([self.exportArray count] > 0) {
                    [self kickOffExportAllNext];
                } else {
                    [NSApp performSelectorOnMainThread:@selector(endSheet:) withObject:self.savingSheet waitUntilDone:NO];
                    [self.savingSheet orderOut: self];
                }
             
                break;
            }
        }
        
    }
}

@end
